mod table_directory;
mod character_map;
mod maximum_profile;
mod font_header;
mod index_to_location;
mod glyph_data;
mod search;

use std::{error::Error, fs::File, io::{BufReader, Read}};

use bincode::Options;
use character_map::CharacterMap;
use font_header::FontHeader;
use glyph_data::GlyphTable;
use index_to_location::{IndexToLocation, IndexToLocationArgs};
use log::{debug, info};
use maximum_profile::MaximumProfile;
use table_directory::TableDirectory;

pub use glyph_data::{GlyphData, GlyphHeader, GlyphPoint, Glyph, SimpleGlyph, SplineElement, SplineElementData};

fn deserialize<R: Read, T: serde::de::DeserializeOwned>(reader: R) -> bincode::Result<T> {
	bincode::options().with_big_endian().with_fixint_encoding().deserialize_from::<R, T>(reader)
}

fn deserialize_vec<R: Read, T: serde::de::DeserializeOwned>(reader: &mut R, entries: usize) -> bincode::Result<Vec<T>> {
	if entries == 0 {
		return Ok(vec![]);
	}

	let mut data: Vec<T> = Vec::with_capacity(entries);

	for _ in 0..entries {
		data.push(deserialize::<_, T>(reader.by_ref())?);
	}

	Ok(data)
}

#[derive(Debug)]
pub struct Font {
	table_directory: TableDirectory,
	font_header: FontHeader,
	character_map: CharacterMap,
	maximum_profile: MaximumProfile,
	index_to_location: IndexToLocation,
	glyph_table: GlyphTable
}
	
impl Font {
	pub fn new(file: File) -> bincode::Result<Self> {
		let mut buf_reader = BufReader::new(file.try_clone()?);
		let table_directory = TableDirectory::new(&mut buf_reader)?;
		debug!("{table_directory:#?}");

		let font_header: FontHeader = table_directory.get_table(&file)?;
		debug!("{font_header:#?}");

		let character_map = table_directory.get_table(&file)?;
		// debug!("{character_map:#?}");

		let maximum_profile: MaximumProfile = table_directory.get_table(&file)?;
		debug!("{maximum_profile:#?}");

		let index_to_location = table_directory.get_table_with_args(&file, Some(IndexToLocationArgs {
			format: font_header.index_to_loc_format,
			num_glyphs: maximum_profile.num_glyphs()
		}))?;

		let glyph_table = table_directory.get_table(&file)?;

		Ok(Font {
			table_directory,
			font_header,
			character_map,
			maximum_profile,
			index_to_location,
			glyph_table
		})
	}

	pub fn get_data_for(&mut self, glyph: char) -> Result<Glyph, Box<dyn Error>> {
		let Some(glyph_id) = self.character_map.get_char(glyph) else {
			return Err("Failed to get glyph id for char".into());
		};

		info!("glyph id for char '{glyph}' is {glyph_id}");

		let glyph = self.glyph_table.get_glyph(glyph_id, &mut self.index_to_location)?;

		Ok(glyph)
	}
}

