use std::{fs::File, io::{BufReader, Read, Seek, SeekFrom}};

use bincode::{ErrorKind, Result};
use log::{debug, error, warn};
use serde::{de::DeserializeOwned, Deserialize};

use super::{deserialize, search::SearchParameters, table_directory::{FontTable, TableDirectoryRecord}};

#[derive(Debug, Deserialize, Copy, Clone)]
struct SegmentToDeltaHeader {
	_length: u16,
	_language: u16,
}

#[derive(Debug)]
struct SegmentToDelta {
	data_source: BufReader<File>,

	header: SegmentToDeltaHeader,
	search_params: SearchParameters,

	end_code: Vec<u16>,
	start_code: Vec<u16>,
	id_delta: Vec<i16>,
	id_range_offset: Vec<u16>,

	glyph_table: u32
}

impl SegmentToDelta {
	fn get_next_vec<R: Read + Seek, T: DeserializeOwned>(reader: &mut R, seg_count: u16) -> Result<Vec<T>> {
		let mut values: Vec<T> = vec!();
		for _ in 0..seg_count {
			values.push(deserialize(reader.by_ref())?);
		}

		Ok(values)
	}

	fn new(mut reader: BufReader<File>) -> Result<SegmentToDelta> {
		let header: SegmentToDeltaHeader = deserialize(reader.by_ref())?;
		let search_params: SearchParameters = deserialize(reader.by_ref())?;

		let end_code = Self::get_next_vec(&mut reader, search_params.seg_count_x2 >> 1)?;
		reader.seek_relative(2)?;
		let start_code = Self::get_next_vec(&mut reader, search_params.seg_count_x2 >> 1)?;
		let id_delta = Self::get_next_vec(&mut reader, search_params.seg_count_x2 >> 1)?;
		let id_range_offset = Self::get_next_vec(&mut reader, search_params.seg_count_x2 >> 1)?;

		let glyph_table_pos = reader.stream_position()? as u32;

		Ok(SegmentToDelta {
			data_source: reader,
			glyph_table: glyph_table_pos,
			header,
			search_params,
			end_code,
			start_code,
			id_delta,
			id_range_offset
		})
	}

	fn get_glyph_id(&mut self, c: u16) -> Option<u16> {
		// Find index of first endcode that is greater or equal to c
		let index =self.search_params.find_first_goe(self.end_code.as_slice(), c)?;
		debug!("First endcode that is >={c} has index {index} (={})", self.end_code[index]);

		if self.start_code[index] > c {
			error!("Start code belonging to this index ({}) is too big", self.start_code[index]);
			return None
		}

		let id_range_offset = self.id_range_offset[index];
		debug!("range offset = {id_range_offset}");

		if id_range_offset == 0 {
			debug!("delta = {}", self.id_delta[index]);
			return Some((c as i16 + self.id_delta[index]) as u16 % u16::MAX);
		}

		let glyph_map_offset = ((id_range_offset >> 1) as usize) + ((c - self.start_code[index]) as usize) - index;
		if let Err(error) = self.data_source.seek(SeekFrom::Start(self.glyph_table as u64 + glyph_map_offset as u64)) {
			error!("{error}");
			return None;
		}

		if let Ok(glyph) = deserialize::<_, u16>(self.data_source.by_ref()) {
			return Some(glyph);
		}

		return None;
	}
}

#[derive(Debug)]
pub struct CharacterMap {
	_table_start: u32,
	_length: u32,

	mapping_table: SegmentToDelta
}

#[derive(Debug, Deserialize)]
struct EncodingRecord {
	platform: u16,
	encoding: u16,
	offset: u32,
}

impl CharacterMap {
	fn find_unicode_table<R: Read + Seek>(reader: &mut R, num_tables: u16) -> Option<EncodingRecord> {
		for _ in 0..num_tables {
			let Ok(table) = deserialize::<_, EncodingRecord>(reader.by_ref()) else {
				return None;
			};

			if table.platform == 0 && table.encoding == 3 {
				return Some(table);
			}
		}

		None
	}
	
	pub fn get_char(&mut self, c: char) -> Option<u16> {
		let code = c as u32;
		if code > (u16::MAX as u32) {
			warn!("Codepoint {code} for '{c}' is not a valid codepoint");
			return None;
		}

		let code = (code & 0xFFFF) as u16;
		debug!("Getting glyph id for codepoint {code}");
		return self.mapping_table.get_glyph_id(code);
	}
}

impl FontTable for CharacterMap {
	const TAG: &'static str = "cmap";
	type InitData = ();

	fn decode(mut reader: BufReader<File>, table: &TableDirectoryRecord, _: Option<Self::InitData>) -> Result<Self> 
		where Self: Sized 
	{
		reader.seek_relative(2)?; // Skip version because i dont care
		let num_tables: u16 = deserialize(reader.by_ref())?;

		let Some(record) = Self::find_unicode_table(&mut reader, num_tables) else {
			return Err(Box::new(ErrorKind::Custom("No encoding for Unicode 2.0 BMP".into())));
		};

		reader.seek(SeekFrom::Start((table.offset + record.offset) as u64))?;
		let format: u16 = deserialize(reader.by_ref())?;

		if format != 4 {
			todo!();
		}

		Ok(CharacterMap {
			mapping_table: SegmentToDelta::new(reader)?,

			_table_start: table.offset,
			_length: table.length,

		})
	}
}
