use std::{collections::HashMap, fs::File, io::{BufReader, Read, Seek, SeekFrom}, iter::{zip, Enumerate, Peekable}, slice::Iter};

use bincode::{ErrorKind, Result};
use bitfield::bitfield;
use log::{debug, error, info, warn};
use serde::{Deserialize, Serialize};

use super::{deserialize, deserialize_vec, index_to_location::IndexToLocation, table_directory::{FontTable, TableDirectoryRecord}};


#[derive(Debug, Deserialize)]
pub struct GlyphHeader {
	num_countours: i16,

	pub xmin: i16,
	pub ymin: i16,
	pub xmax: i16,
	pub ymax: i16
}

bitfield! {
	#[derive(Copy, Clone, Deserialize)]
	struct GlyphFlagData(u8);
	impl Debug;
	on_curve, _: 0;
	x_short, _: 1;
	y_short, _: 2;
	repeat, _: 3;
	x_same_or_positive, _: 4;
	y_same_or_positive, _: 5;
	overlap_simple, _: 6;
	reserved, _: 7;
}

#[derive(Debug)]
struct GlyphFlag {
	flags: GlyphFlagData,
	repeat: u8
}

#[derive(Debug)]
struct GlyphFlags<'a> {
	flags: &'a Vec<GlyphFlag>,
	current_item: usize,
	repetitions: u16
}

impl<'a> GlyphFlags<'a> {
	fn new(flags: &'a Vec<GlyphFlag>) -> GlyphFlags<'a> {
		GlyphFlags { flags, current_item: 0, repetitions: 0 }
	}
}

impl<'a> Iterator for GlyphFlags<'a> {
	type Item = GlyphFlagData;

	fn next(&mut self) -> Option<Self::Item> {
		if self.current_item >= self.flags.len() {
			debug!("no more logical flags");
			return None;
		}

		let current = &self.flags[self.current_item];

		if self.repetitions <= current.repeat as u16 {
			if self.repetitions > 0 {
				debug!("flag #{} repeated {} times", self.current_item, self.repetitions);
			} else {
				debug!("yield flag #{}", self.current_item);
			}

			self.repetitions += 1;
			return Some(current.flags);
		}

		self.repetitions = 0;
		self.current_item += 1;
		self.next()
	}
}

#[derive(Debug, Copy, Clone, Serialize)]
pub struct GlyphPoint {
	pub x: i32,
	pub y: i32,

	pub on_curve: bool,
	pub is_virtual: bool,
	pub is_endpoint: bool,
}

#[derive(Debug)]
pub struct SimpleGlyph {
	instructions: Vec<u8>,

	points: Vec<GlyphPoint>
}

impl SimpleGlyph {
	fn read_flags<R: Read + Seek>(reader: &mut R, num_points: usize) -> Result<Vec<GlyphFlag>> {
		debug!("reading flags from position {}", reader.stream_position()?);
		let mut flags: Vec<GlyphFlag> = vec![];

		let mut points_processed: usize = 0;
		while points_processed <= num_points {
			let flag: GlyphFlagData = deserialize(reader.by_ref())?;

			let repetitions: u8 = if flag.repeat() { deserialize(reader.by_ref())? } else { 0 };
			flags.push(GlyphFlag {
				flags: flag,
				repeat: repetitions
			});

			points_processed += repetitions as usize + 1;
			debug!("Flag: {flag:#?}, repeated {repetitions} times, logical elements processed: {points_processed}")
		}

		// debug!("glyph flags: {flags:#?}");
		Ok(flags)
	}

	fn read_coordinates<R: Read + Seek>(reader: &mut R, flags: GlyphFlags, is_y: bool) -> Result<Vec<(i32, bool)>> {
		debug!("reading batch of {} coordinates at {}", if is_y { "y" } else { "x" }, reader.stream_position()?);
		let mut coordinates: Vec<(i32, bool)> = vec![];
		let mut current_coordinate = 0i32;

		for flag in flags {
			debug!("reading next coordinate");

			let short_vec = if is_y { flag.y_short() } else { flag.x_short() };
			let same_or_positive = if is_y { flag.y_same_or_positive() } else { flag.x_same_or_positive() };

			debug!("short vec: {short_vec}, same or positive: {same_or_positive}");

			if short_vec {
				let sign: i32 = if same_or_positive { 1 } else { -1 };
				let delta = sign * (deserialize::<_, u8>(reader.by_ref())? as i32);

				debug!("delta = (u8){delta}");

				current_coordinate += delta;
				coordinates.push((current_coordinate, flag.on_curve()));
			} else {
				if same_or_positive {
					debug!("coordinate is repeated");
					coordinates.push((current_coordinate, flag.on_curve()));
				} else {
					let delta: i16 = deserialize(reader.by_ref())?;
					debug!("delta = (i16){delta}");
					current_coordinate += delta as i32;
					coordinates.push((current_coordinate, flag.on_curve()));
				}
			}

			debug!("current coordinate is now {current_coordinate}");
		}

		Ok(coordinates)
	}

	fn new<R: Read + Seek>(header: &GlyphHeader, reader: &mut R) -> Result<SimpleGlyph> {
		let end_points: Vec<u16> = deserialize_vec(reader, header.num_countours as usize)?;
		debug!("glyph end points: {end_points:#?}");

		let num_instructions: u16 = deserialize(reader.by_ref())?;
		let instructions: Vec<u8> = deserialize_vec(reader, num_instructions as usize)?;
		debug!("glyph instructions: {num_instructions}");

		let flags = SimpleGlyph::read_flags(reader, *end_points.last().unwrap() as usize)?;

		let x_coords = SimpleGlyph::read_coordinates(reader, GlyphFlags::new(&flags), false)?;
		let y_coords = SimpleGlyph::read_coordinates(reader, GlyphFlags::new(&flags), true)?;

		if x_coords.len() != *end_points.last().unwrap() as usize + 1 {
			warn!("expected {} x coordinates, got {} instead", end_points.last().unwrap() + 1, x_coords.len());
		}

		if y_coords.len() != *end_points.last().unwrap() as usize + 1 {
			warn!("expected {} y coordinates, got {} instead", end_points.last().unwrap() + 1, y_coords.len());
		}

		let mut current_point: u16 = 0;
		let coordinates = zip(x_coords.iter().cloned(), y_coords.iter().cloned())
			.map(|(x, y)| {
				let point = GlyphPoint {
					x: x.0, y: y.0, on_curve: x.1, is_endpoint: end_points.contains(&current_point), is_virtual: false
				};

				current_point += 1;

				point
			});

		// coordinates.for_each(|point| debug!("({}, {}, {})", point.x, point.y, if point.on_curve { "on" } else { "off" }));

		Ok(SimpleGlyph { instructions, points: coordinates.collect() })
	}

	pub fn iter<'a>(&'a self) -> SimpleGlyphIter<'a> {
		SimpleGlyphIter {
			glyph: self,
			last_point: None,
			iter: self.points.iter().peekable()
		}
	}
}

struct SimpleGlyphIter<'a> {
	glyph: &'a SimpleGlyph,
	last_point: Option<GlyphPoint>,
	iter: Peekable<Iter<'a, GlyphPoint>>
}

impl<'a> Iterator for SimpleGlyphIter<'a> {
	type Item = GlyphPoint;

	fn next(&mut self) -> Option<Self::Item> {
		if let Some(last_point) = self.last_point {
			let Some(&next_point) = self.iter.peek() else {
				return None
			};

			if !last_point.on_curve && !next_point.on_curve {
				// Interpolate a point
				let virtual_point = GlyphPoint {
					x: last_point.x + (next_point.x - last_point.x) / 2,
					y: last_point.y + (next_point.y - last_point.y) / 2,
					is_endpoint: false,
					is_virtual: true,
					on_curve: true
				};

				self.last_point = Some(virtual_point);
				return self.last_point;
			}
		}

		self.last_point = self.iter.next().copied();
		self.last_point
	}
}

#[derive(Debug)]
pub enum GlyphData {
	Simple(SimpleGlyph),
	Composite
}

#[derive(Debug)]
pub struct Glyph {
	pub header: GlyphHeader,
	pub data: GlyphData
}

#[derive(Debug, Clone, Copy)]
pub struct IndexedGlyphPoint {
	pub point: GlyphPoint,
	pub index: usize
}

#[derive(Debug)]
pub enum SplineElementData {
	Line(IndexedGlyphPoint, IndexedGlyphPoint),
	Bezier(IndexedGlyphPoint, IndexedGlyphPoint, IndexedGlyphPoint)
}

#[derive(Debug)]
pub struct SplineElement {
	pub data: SplineElementData,
	pub is_last: bool
}

pub struct SplineIter<'a> {
	points: Enumerate<Box<dyn Iterator<Item = GlyphPoint> + 'a>>,

	last_point: Option<IndexedGlyphPoint>,
	final_control: Option<IndexedGlyphPoint>,
	first_point: Option<IndexedGlyphPoint>
}

impl<'a> SplineIter<'a> {
	fn handle_spline_start(&mut self) -> Result<()> {
		let Some((index, point)) = self.points.next() else {
			return Err(ErrorKind::Custom("Not enough points in contous to form a spline".into()).into());
		};

		if point.on_curve {
			self.last_point = Some(IndexedGlyphPoint { point, index });
			self.first_point = Some(IndexedGlyphPoint { point, index });
			self.final_control = None;
			return Ok(());
		} else {
			self.final_control = Some(IndexedGlyphPoint { point, index });
		}

		let Some((index, point)) = self.points.next() else {
			return Err(ErrorKind::Custom("Not enough points in contous to form a spline".into()).into());
		};

		if !point.on_curve {
			return Err(ErrorKind::Custom("Iterator returned invalid point sequence".into()).into());
		}

		self.last_point = Some(IndexedGlyphPoint { point, index });
		self.first_point = Some(IndexedGlyphPoint { point, index });

		Ok(())
	}

	fn handle_spline_close(&mut self) -> Option<SplineElement> {
		let Some(last_point) = self.last_point else {
			return None;
		};

		if let Some(final_control) = self.final_control {
			Some(SplineElement {
				data: SplineElementData::Bezier(last_point, final_control, self.first_point.unwrap()),
				is_last: true
			})
		} else {
			Some(SplineElement {
				data: SplineElementData::Line(last_point, self.first_point.unwrap()),
				is_last: true
			})
		}
	}
}

impl<'a> Iterator for SplineIter<'a> {
	type Item = SplineElement;

	fn next(&mut self) -> Option<Self::Item> {
		let Some(last_point) = self.last_point else {
			debug!("spline iter: last point is None");
			return None;
		};

		if last_point.point.is_endpoint {
			debug!("spline iter: point is an endpoint of contour. closing spline, starting new one (if possible)");
			let result = self.handle_spline_close();
			if let Err(err) = self.handle_spline_start() {
				debug!("spline iter: no more contours");
				self.last_point = None;
			}

			return result;
		}

		let Some((index, point)) = self.points.next() else {
			self.last_point = None;
			debug!("spline iter: no more points, closing spline");
			return self.handle_spline_close();
		};

		if point.on_curve {
			self.last_point = Some(IndexedGlyphPoint { point, index });
			debug!("spline iter: point is on curve");

			return Some(SplineElement {
				data: SplineElementData::Line(last_point, IndexedGlyphPoint { point, index }),
				is_last: false
			});
		} 

		debug!("spline iter: point is control point");

		let Some((next_index, next_point)) = self.points.next() else {
			debug!("spline iter: no more data after control point, closing spline");
			self.last_point = None;
			return self.handle_spline_close();
		};

		debug!("spline iter: returning bezier");
		self.last_point = Some(IndexedGlyphPoint { point: next_point, index: next_index });
		Some(SplineElement {
			data: SplineElementData::Bezier(last_point, IndexedGlyphPoint { point, index }, IndexedGlyphPoint { point: next_point, index: next_index }),
			is_last: false
		})
	}
}

impl Glyph {
	pub fn iter<'a>(&'a self) -> Box<dyn Iterator<Item = GlyphPoint> + 'a> {
		Box::new(
			match &self.data {
				GlyphData::Simple(glyph) => glyph.iter(),
				GlyphData::Composite => todo!()
			}
		)
	}

	pub fn splines<'a>(&'a self) -> Result<SplineIter<'a>> {
		let mut splines = SplineIter {
			points: self.iter().enumerate(),
			last_point: None,
			final_control: None,
			first_point: None
		};

		splines.handle_spline_start()?;

		Ok(splines)
	}
}

#[derive(Debug)]
pub struct GlyphTable {
	data_source: BufReader<File>,
	offset: u32,

	cache: HashMap<u16, Glyph>
}

impl GlyphTable {
	pub fn get_glyph(&mut self, glyph_id: u16, location: &mut IndexToLocation) -> Result<Glyph>{
		let Some(offset) = location.get_offset(glyph_id) else {
			return Err(ErrorKind::Custom("Failed to look up glyph location".into()).into());
		};

		debug!("glyph has offset {offset}");
		self.data_source.seek(SeekFrom::Start(self.offset as u64 + offset as u64))?;

		debug!("reading glyph data from position {}", self.offset as u64 + offset as u64);

		let header: GlyphHeader = deserialize(self.data_source.get_ref())?;
		debug!("glyph header: {header:#?}");

		let result = if header.num_countours >= 0 {
			Ok(Glyph {
				data: GlyphData::Simple(SimpleGlyph::new(&header, self.data_source.by_ref())?),
				header
			})
		} else {
			todo!();
		};

		// sanity check
		let stream_pos = self.data_source.stream_position()?;
		let Some(next_offset) = location.get_offset(glyph_id + 1) else {
			warn!("Failed to perform sanity check after glyph parsing");
			return result;
		};

		let remaining_bytes = stream_pos as i64 - (self.offset + next_offset) as i64;
		if remaining_bytes != 0 {
			warn!("sanity check failed, there are still {remaining_bytes} bytes remaining in this glyph entry (read head: {}, table off: {}, location: [{offset}, {next_offset}])", stream_pos, self.offset);
		}

		result
	}
}

impl FontTable for GlyphTable {
	const TAG: &'static str = "glyf";
	type InitData = ();

	fn decode(reader: BufReader<File>, table: &TableDirectoryRecord, _: Option<Self::InitData>) -> bincode::Result<Self>
		where Self: Sized 
	{
		Ok(GlyphTable {
			data_source: reader,
			offset: table.offset,

			cache: HashMap::new()
		})
	}
}
