use std::{collections::HashMap, fs::File, io::{BufReader, Seek, SeekFrom}};

use log::{debug, error};
use serde::de::DeserializeOwned;

use crate::font::deserialize;

use super::table_directory::{FontTable, TableDirectoryRecord};


#[derive(Debug)]
pub struct IndexToLocation {
	data_source: BufReader<File>,
	offset: u32,

	num_glyphs: u16,
	stride: u16,

	cache: HashMap<u16, u32>
}

impl IndexToLocation {
	fn read_offset_with_stride<T: DeserializeOwned>(&self) -> Option<u32> 
		where u32: From<T>
	{
		match deserialize::<_, T>(self.data_source.get_ref()) {
			Ok(val) => Some(u32::from(val)),
			Err(err) => {
				error!("{err}");
				None
			}
		}
	}

	fn read_offset(&self) -> Option<u32> {
		match self.stride {
			2 => self.read_offset_with_stride::<u16>(),
			4 => self.read_offset_with_stride::<u32>(),
			_ => None
		}
	}

	pub fn get_offset(&mut self, glyph_id: u16) -> Option<u32> {
		if glyph_id > self.num_glyphs {
			error!("glyph id {glyph_id} out of range");
			return None;
		}

		if let Some(value) = self.cache.get(&glyph_id) {
			debug!("Found glyph {glyph_id} in cache");
			return Some(*value);
		}

		if let Err(err) = self.data_source.seek(SeekFrom::Start(self.offset as u64 + (glyph_id * self.stride) as u64)) {
			error!("{err}");
			return None;
		}

		debug!("Reading glyph offset from file");

		let offset = self.read_offset()?;
		self.cache.insert(glyph_id, offset);

		Some(offset)
	}
}

pub struct IndexToLocationArgs {
	pub format: i16,
	pub num_glyphs: u16
}

impl FontTable for IndexToLocation {
	const TAG: &'static str = "loca";
	type InitData = IndexToLocationArgs;

	fn decode(reader: BufReader<File>, table: &TableDirectoryRecord, args: Option<Self::InitData>) -> bincode::Result<Self>
		where Self: Sized
	{
		let args = args.unwrap();
		if args.format > 1 {
			return Err(bincode::ErrorKind::Custom(format!("Invalid index to location format: '{}'", args.format)).into());
		}

		Ok(IndexToLocation {
			data_source: reader,
			offset: table.offset,

			stride: 2 + (args.format as u16) * 2,
			num_glyphs: args.num_glyphs,

			cache: HashMap::new()
		})
	}
}
