use std::{collections::HashMap, fs::File, io::{BufReader, Read, Seek, SeekFrom}};

use bincode::Result;
use log::debug;
use serde::Deserialize;

use crate::font::deserialize;

#[derive(Deserialize, Debug)]
struct TableDirectoryHeader {
	sfnt_version: u32,
	num_tables: u16,
}

#[derive(Deserialize, Debug, Copy, Clone)]
pub struct TableDirectoryRecord {
	pub table_tag: u32,
	pub checksum: u32,
	pub offset: u32,
	pub length: u32
}

#[derive(Debug, Default)]
pub struct TableDirectory(pub HashMap<u32, TableDirectoryRecord>);

impl TableDirectory {
	pub fn new<R: Read + Seek>(reader: &mut R) -> Result<TableDirectory> {
		let header: TableDirectoryHeader = deserialize(reader.by_ref())?;
		debug!("{header:#x?}");
		let mut tables: HashMap<u32, TableDirectoryRecord> = HashMap::with_capacity(header.num_tables as usize);

		reader.seek_relative(3 * size_of::<u16>() as i64)?;
		
		for _ in 0..header.num_tables {
			let record: TableDirectoryRecord = deserialize(reader.by_ref())?;
			tables.insert(record.table_tag, record);
		}

		Ok(TableDirectory(tables))
	}

	pub fn get_table<T: FontTable>(&self, file: &File) -> Result<T> {
		self.get_table_with_args(file, None)
	}

	pub fn get_table_with_args<T: FontTable>(&self, file: &File, init_data: Option<T::InitData>) -> Result<T> {
		let table_id = T::TAG.as_bytes().first_chunk::<4>().unwrap().iter().fold(0u32, |running, &val| {
			(running << 8) | (val as u32)
		});

		T::find_and_decode(file, self.0.get(&table_id).ok_or(bincode::ErrorKind::Custom(format!("No table with id '{}' found", T::TAG)))?, init_data)
	}

}

pub trait FontTable {
	const TAG: &'static str;
	type InitData;

	fn decode(reader: BufReader<File>, table: &TableDirectoryRecord, init_data: Option<Self::InitData>) -> Result<Self>
		where Self: Sized;

	fn find_and_decode(file: &File, table: &TableDirectoryRecord, init_data: Option<Self::InitData>) -> Result<Self> 
		where Self: Sized
	{
		let file = file.try_clone()?;
		let mut reader = BufReader::new(file);

		reader.seek(SeekFrom::Start(table.offset as u64))?;
		Self::decode(reader, table, init_data)
	}
}
