#[macro_use] extern crate rocket;

use std::{fs::File, future::Future, net::{IpAddr, Ipv4Addr}, sync::Mutex};

use fontloader::{writer::Visitor, JsonWriter, Font};
use rocket::{fairing::{Fairing, Info, Kind}, http::Header, response::{content, status}, Config, State};

struct CORS;

#[rocket::async_trait]
impl Fairing for CORS {
	fn info(&self) -> Info {
		Info {
			name: "Adding CORS headers to responses",
			kind: Kind::Response
		}
	}

	async fn on_response<'r>(&self, _req: &'r rocket::Request<'_>, response: &mut rocket::Response<'r>) {
		response.set_header(Header::new("Access-Control-Allow-Origin", "*"));
        response.set_header(Header::new("Access-Control-Allow-Methods", "POST, GET, PATCH, OPTIONS"));
        response.set_header(Header::new("Access-Control-Allow-Headers", "*"));
        response.set_header(Header::new("Access-Control-Allow-Credentials", "true"));
	}
}

struct SharedFont {
	font: Mutex<Font>
}

#[get("/glyph/<glyph>")]
fn get_glyph(font: &State<SharedFont>, glyph: &str) -> status::Accepted<content::RawJson<String>> {
	let Some(char) = glyph.chars().next() else {
		return status::Accepted(content::RawJson("".into()));
	};

	let result = match font.font.lock().unwrap().get_data_for(char) {
		Ok(result) => result,
		Err(err) => return status::Accepted(content::RawJson(err.to_string()))
	};

	let mut json_buf = vec![];
	JsonWriter::new(&mut json_buf).write(&result);

	status::Accepted(content::RawJson(String::from_utf8(json_buf).unwrap()))
}

#[launch]
fn rocket() -> _ {
	let _ = fontloader::init();

	let font = Font::new(File::open("/fonts/Helvetica.ttf").unwrap());
	let Ok(font) = font else {
        let err = font.unwrap_err().to_string();
        panic!("{err}");
	};

	let mut config = Config::release_default();
	config.address = IpAddr::V4(Ipv4Addr::new(0, 0, 0, 0));
	config.port = 8000;

	rocket::custom(config)
		.manage(SharedFont { font: Mutex::from(font) })
		.mount("/", routes![get_glyph])
		.attach(CORS)
}
