use std::io::Write;

use log::debug;
use serde::Serialize;

use crate::font::{GlyphHeader, GlyphPoint, SplineElement, SplineElementData};

use super::Visitor;

#[derive(Serialize)]
enum ContourElement {
	Line(usize, usize),
	Bezier(usize, usize, usize)
}

#[derive(Serialize)]
struct Contour {
	points: Vec<GlyphPoint>,
	elements: Vec<ContourElement>
}

#[derive(Serialize)]
struct BoundingBox {
	xmin: i16,
	xmax: i16,
	ymin: i16,
	ymax: i16
}

#[derive(Serialize)]
struct JsonData {
	bounding_box: BoundingBox,
	contours: Vec<Contour>
	
}

pub struct JsonWriter<W: Sized + Write> {
	writer: W,

	index_correction: usize,
	current_contour: usize,

	body: JsonData
}

impl<W: Sized + Write> JsonWriter<W> {
	pub fn new(writer: W) -> JsonWriter<W> {
		JsonWriter {
			writer,
			index_correction: 0,
			current_contour: 0,

			body: JsonData {
				bounding_box: BoundingBox { xmin: 0, xmax: 0, ymin: 0, ymax: 0 },
				contours: vec![]
			}
		}
	}
}

impl<W: Sized + Write> Visitor for JsonWriter<W> {
	fn write_prefix(&mut self, header: &GlyphHeader) {
		self.body.bounding_box = BoundingBox {
			xmin: header.xmin,
			xmax: header.xmax,
			ymin: header.ymin,
			ymax: header.ymax
		};

		self.body.contours.push(Contour {
			points: vec![],
			elements: vec![]
		});
	}

	fn write_point(&mut self, point: &GlyphPoint) {
		self.body.contours.last_mut().unwrap().points.push(*point);

		if point.is_endpoint {
			self.body.contours.push(Contour {
				points: vec![],
				elements: vec![]
			})
		}

	}

	fn write_spline(&mut self, spline: &SplineElement) {
		debug!("index correction: {}", self.index_correction);
		let indexed_spline = match spline.data {
			SplineElementData::Line(start, end) => ContourElement::Line(start.index - self.index_correction, end.index - self.index_correction),
			SplineElementData::Bezier(start, control, end) => ContourElement::Bezier(start.index - self.index_correction, control.index - self.index_correction, end.index - self.index_correction)
		};

		debug!("pushing spline to contours array at position {}/{}", self.current_contour, self.body.contours.len());
		self.body.contours[self.current_contour].elements.push(indexed_spline);
		if spline.is_last {
			self.index_correction = self.body.contours[self.current_contour].points.len();
			self.current_contour += 1;
		}
	}

	fn write_suffix(self) {
		serde_json::to_writer(self.writer, &self.body);
	}
}
